import datetime
import sys
import traceback

import uuid
import os.path
from typing import List, Dict, Union
import time
import random
import threading
from pychatgpt import Chat, Options
from pychatgpt.classes.chat import ask
from fastchatgpt.chatgpt.fast_chat import FastChat
import copy
from tqdm import tqdm


def parse_proxy(proxies:Union[Dict, str, None]) -> Union[Dict, None]:
    if isinstance(proxies, str):
        parts = proxies.split('+')
        _proxies = {}
        for part in parts:
            if part.startswith('http_proxy='):
                _proxies['http'] = part.split('http_proxy=')[1]
            if part.startswith('https_proxy='):
                _proxies['https'] = part.split('https_proxy=')[1]
        return _proxies
    return proxies


class TokenChat:
    """
    用来适配通过access_token来获取的情况

    """
    def __init__(self, access_token, proxies):
        # Manually set the token
        # OpenAI.Auth.save_access_token(access_token=access_token, expiry=time.time() + 3600)

        # Get the token, expiry
        # access_token, expiry = OpenAI.Auth.get_access_token()
        self.access_token = (access_token, time.time()+3600)
        self.proxies = proxies

    def ask(self, prompt):
        answer, previous_convo, convo_id = ask(auth_token=self.access_token,
                prompt=prompt,
                conversation_id=None,
                previous_convo_id=None,
                proxies=self.proxies)
        return answer


def get_account(accounts)->List[Dict]:
    """
    Read accounts to a list of dict

    :param accounts:
    :return:
    """
    if isinstance(accounts, str):
        _account = []
        assert os.path.exists(accounts), 'Please pass an account file when you pass accounts as a string'
        with open(accounts, 'r') as f:
            for line in f:
                parts = line.strip().split()
                if len(parts) == 1:  # access token
                    _account.append({
                        'access_token': parts[0]
                    })
                else:
                    _account.append({
                        'username': parts[0],
                        'password': parts[1]
                    })
                    if len(parts)>2:
                        proxy = parse_proxy(parts[2])
                        _account[-1]['proxies'] = proxy

        return _account
    return accounts


class Balancer:
    """
    Balancer is used to balance different accounts, so that we can call Bot concurrently.

    """
    def __init__(self, accounts:Union[List[Dict], str], proxies:Union[Dict, str, None],
                 each_query_time_sep:int=30):
        """

        :param accounts:
        :param proxies:
        :param each_query_time_sep:
        """
        self.accounts = get_account(accounts)
        self.proxies = parse_proxy(proxies)
        self.chats = {}
        self.login_ins()
        self.spare_chats = {k:(v, time.time()) for k,v in self.chats.items()}
        self.chat_last_send = {k:-1 for k in self.chats}
        self.each_query_time_sep = each_query_time_sep

    def login_ins(self):
        fail_accounts = []
        for account in self.accounts:
            option = Options()
            if self.proxies:
                option.proxies = self.proxies
            if account.get('proxies', None):
                option.proxies = account.get('proxies')
            print(f"Login to {account['username']}...")
            try:
                if 'access_token' in account:
                    chat = TokenChat(access_token=account['access_token'], proxies=self.proxies)
                    self.chats[account['username']] = chat
                else:
                    chat = FastChat(email=account['username'], password=account['password'], options=option)
                    # import pdb
                    # pdb.set_trace()
                    if chat.auth_access_token is not None:
                        self.chats[account['username']] = chat
                        print(f"Login to {account['username']} Successfully...")
                    else:
                        print(f"Fail to login to {account['username']}...")
            except Exception as e:
                fail_accounts.append(account['username'])
                print(f"Fail to login to {account['username']}...")
                import traceback
                traceback.print_exc()
            # sleep several seconds
            time.sleep(random.randint(100, 400)/100)
        if len(self.chats)==0:
            raise RuntimeError("No success login...")
        print(f"Successful log {len(self.chats)} chats...")
        print(f"Fail to login to {fail_accounts}")

    def batch_query(self, prompts:List[str]) -> List[str]:
        pass


class SimpleBalancer(Balancer):
    def __init__(self, accounts:Union[List[Dict], str], proxies:Union[Dict, str, None]=None,
                 refresh_dialogue_each_prompt=True, each_query_time_sep=30):
        """
        在不同账号之间进行轮询的并行 Balancer，支持传入多个账号，会考虑到当出现了禁止访问后让对应的账号暂停访问1小时。Balancer会首先尝试自动
        登录所有的账号，然后通过这些账号进行轮询访问。

        :param accounts: (1) str类型，表明该参数为一个文件路径，文件中的内容以每行为单位，每一行以空格分割，第一个元素为账号，第二个元素为
            密码，第三个元素（可选）为当前账号使用的代理，代理的格式需要类似http_proxy=http://proxy.com+https_proxy=http://proxy.com；
            (2) List[Dict]，类似[{'username': xxx, 'pasdword': xxx, 'proxies': xxx},
                {'username': xxx, 'pasdword': xxx, 'proxies': xxx}]其中proxies是可选参数
        :param proxies: （1）Dict, 类似{'http_proxy': 'http://proxy.com', 'https_proxy': 'http://proxy.com'}
            (2) str，类似http_proxy=http://proxy.com+https_proxy=http://proxy.com；
        :param refresh_dialogue_each_prompt: 是否在每次请求后都刷新对话状态，通过这个方式可以使得不同 sample 之间没有影响
        :param each_query_time_sep: 每个账号的连续两次访问间间隔的时间
        """
        super().__init__(accounts=accounts, proxies=proxies, each_query_time_sep=each_query_time_sep)
        self.refresh_dialogue_each_prompt = refresh_dialogue_each_prompt

    def batch_query(self, prompts:Dict[int, Dict]) -> Dict:
        """

        :param prompts: Each item in this dict should include an id and its corresponding text. The id will be used to check
            whether this text has been annotated.
        :return: the return dict has the folloing form
            {
                idx1: {  # from the input prompts
                    'prompt': '',  # from the input prompts
                    'response': ''  # answered by the bot
                    'username': ''  # 这个请求是由哪个账号发出的,
                    'send_timestamp': '' # 发送请求回应的时间
                    'receive_timestamp': '' # 收到请求回应的时间
                },
                idx2: {  # from the input prompts
                    'prompt': '',  # from the input prompts
                    'response': ''  # answered by the bot
                    'username': ''  # 这个请求是由哪个账号发出的,
                    'send_timestamp': '' # 发送请求回应的时间
                    'receive_timestamp': '' # 收到请求回应的时间
                }
            }
        """
        break_seconds = max(self.each_query_time_sep//len(self.chats), 2)
        self.responses = {}
        ts = []
        left_prompts = [(idx, pmt) for idx, pmt in prompts.items()]
        start_time =time.time()
        try:
            with tqdm(total=len(left_prompts)) as pbar:
                while len(left_prompts):
                    idx, prompt = left_prompts.pop()
                    chat_id, chat = self.get_spare_chat()
                    # add random seconds to separate two neighboring request
                    rand_secs = random.randrange(break_seconds//2*100, break_seconds*100)/100
                    time.sleep(rand_secs)
                    # send request
                    t = threading.Thread(target=self.send_query, args=(idx, prompt, chat, chat_id, left_prompts))
                    t.start()
                    ts.append(t)
                    pbar.update(1)

            for t in ts:
                t.join()
        except:
            import traceback
            traceback.print_exc()
        finally:
            print(f"Annotate {len(self.responses)} samples, left {len(left_prompts)} samples, "
                  f"takes {time.time()-start_time} seconds")

        return copy.deepcopy(self.responses)

    def refresh_chat_dialogue(self, chat: Chat):
        """
        重置 Chat 的状态。

        :param chat:
        :return:
        """
        chat.conversation_id = None
        chat.previous_convo_id = str(uuid.uuid4())

    def send_query(self, idx:int, prompt:Dict, chat: Chat, chat_id:int, left_prompts:List):
        """
        实际发送请求的函数

        :param idx: 当前 prompt 的index
        :param prompt: 一个包含了 'prompt' key 的字典
        :param chat: 一个ChatGPT账号对象
        :param chat_id: chat 对应的ID
        :param left_prompts: 还没有请求过的 prompt 内容，如果当前访问失败了，会将这个 prompt 重新加入到这个里面
        :return:
        """
        success = False
        try:
            send_timestamp = str(datetime.datetime.now())
            response = chat.ask(prompt=prompt['prompt'])
            if self.refresh_dialogue_each_prompt:
                self.refresh_chat_dialogue(chat)
            if '[Response Text]' in response:  # 大概率是出问题了
                print(f"Fail to get response:{response}")
            else:
                success = True
                # print(f"Success ask with response:{response} for {chat_id}")
            self.chat_last_send[chat_id] = time.time()
        except:
            import traceback
            traceback.print_exc()
            success = False
        if success:
            self.spare_chats[chat_id] = (chat,
                                         time.time()+random.randint(self.each_query_time_sep//2, self.each_query_time_sep))
            res = copy.deepcopy(prompt)
            res['response'] = response
            if chat.email is not None:
                res['email'] = chat.email
            res['receive_timestamp'] = str(datetime.datetime.now())
            res['send_timestamp'] = send_timestamp
            self.responses[idx] = res
        else:
            # If there is a problem with this response, better to make the agent to wait...
            self.spare_chats[chat_id] = (chat, time.time()+random.randint(3600, 3800))
            left_prompts.append((idx, prompt))
            return None

    def get_spare_chat(self):
        # 找到空闲的chat
        count = 0
        found_spare_chat = False
        while len(self.spare_chats)==0 or found_spare_chat is False:
            time.sleep(1)
            count += 1
            if count>4000:
                raise RuntimeError("It takes too long to retrieve a sentence...")
            for idx in self.spare_chats:
                chat, use_after = self.spare_chats[idx]
                if time.time()>use_after:
                    chat_id, chat = idx, chat
                    found_spare_chat = True
                    break
        # 被选中后，让它不能再被选择
        self.spare_chats[chat_id] = (chat, float('inf'))
        return chat_id, chat



# adapted from https://github.com/rawandahmad698/PyChatGPT

# Builtins
import json
import os
import time
import urllib.parse
from io import BytesIO
import re
import base64
from typing import Tuple

# Client (Thank you!.. https://github.com/FlorianREGAZ)
import tls_client

# BeautifulSoup
from bs4 import BeautifulSoup

# Svg lib
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPM

# Local
from pychatgpt.classes import exceptions as Exceptions

# Fancy stuff
import colorama
from colorama import Fore

colorama.init(autoreset=True)


class Auth:
    def __init__(self, email_address: str, password: str, proxy: str = None):
        self.email_address = email_address
        self.password = password
        self.proxy = proxy
        self.__session = tls_client.Session(
            client_identifier="chrome_105"
        )

    @staticmethod
    def _url_encode(string: str) -> str:
        """
        URL encode a string
        :param string:
        :return:
        """
        return urllib.parse.quote(string)

    def create_token(self):
        """
            Begin the auth process
        """
        if not self.email_address or not self.password:
            print(f"{Fore.RED}[OpenAI] {Fore.WHITE}Please provide an email address and password")
            raise Exceptions.PyChatGPTException("Please provide an email address and password")
        else:
            # Print email address and password
            print(f"{Fore.GREEN}[OpenAI] {Fore.WHITE}Email address: {self.email_address}")
            # Show 3 characters of password, then hide the rest
            print(f"{Fore.GREEN}[OpenAI] {Fore.WHITE}Password: {self.password[:3]}{'*' * len(self.password[3:])}")

            if self.proxy is not None:
                if isinstance(self.proxy, str):
                    proxies = {
                        "http": self.proxy,
                        "https": self.proxy
                    }
                else:
                    proxies = self.proxy
                print(f"{Fore.GREEN}[OpenAI] {Fore.WHITE}Using proxy: {self.proxy}")
                self.__session.proxies = proxies

        print(f"{Fore.GREEN}[OpenAI] {Fore.WHITE}Beginning auth process")
        # First, make a request to https://chat.openai.com/auth/login
        url = "https://chat.openai.com/auth/login"
        headers = {
            "Host": "ask.openai.com",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
        }
        print(f"{Fore.GREEN}[OpenAI][1] {Fore.WHITE}Making request to {url}")

        response = self.__session.get(url=url, headers=headers)
        if response.status_code == 200:
            print(f"{Fore.GREEN}[OpenAI][1] {Fore.WHITE}Request was " + Fore.GREEN + "successful")
            self._part_two()
        else:
            raise Exceptions.Auth0Exception("Failed to make the first request, Try that again!")

    def _part_two(self):
        """
        In part two, We make a request to https://chat.openai.com/api/auth/csrf and grab a fresh csrf token
        """

        print(f"{Fore.GREEN}[OpenAI][2] {Fore.WHITE}Beginning part two")
        url = "https://chat.openai.com/api/auth/csrf"
        headers = {
            "Host": "ask.openai.com",
            "Accept": "*/*",
            "Connection": "keep-alive",
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
            "Referer": "https://chat.openai.com/auth/login",
            "Accept-Encoding": "gzip, deflate, br",
        }
        print(f"{Fore.GREEN}[OpenAI][2] {Fore.WHITE}Grabbing CSRF token from {url}")
        response = self.__session.get(url=url, headers=headers)
        if response.status_code == 200 and 'json' in response.headers['Content-Type']:
            print(f"{Fore.GREEN}[OpenAI][2] {Fore.WHITE}Request was " + Fore.GREEN + "successful")
            csrf_token = response.json()["csrfToken"]
            print(f"{Fore.GREEN}[OpenAI][2] {Fore.WHITE}CSRF Token: {csrf_token}")
            self._part_three(token=csrf_token)
        else:
            raise Exceptions.Auth0Exception("[OpenAI][2] Failed to make the request, Try that again!")

    def _part_three(self, token: str):
        """
        We reuse the token from part to make a request to /api/auth/signin/auth0?prompt=login
        """
        print(f"{Fore.GREEN}[OpenAI][3] {Fore.WHITE}Beginning part three")
        url = "https://chat.openai.com/api/auth/signin/auth0?prompt=login"

        payload = f'callbackUrl=%2F&csrfToken={token}&json=true'
        headers = {
            'Host': 'ask.openai.com',
            'Origin': 'https://chat.openai.com',
            'Connection': 'keep-alive',
            'Accept': '*/*',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Referer': 'https://chat.openai.com/auth/login',
            'Content-Length': '100',
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        print(f"{Fore.GREEN}[OpenAI][3] {Fore.WHITE}Making request to {url}")
        response = self.__session.post(url=url, headers=headers, data=payload)
        if response.status_code == 200 and 'json' in response.headers['Content-Type']:
            print(f"{Fore.GREEN}[OpenAI][3] {Fore.WHITE}Request was " + Fore.GREEN + "successful")
            url = response.json()["url"]
            if url == "https://chat.openai.com/api/auth/error?error=OAuthSignin" or 'error' in url:
                print(f"{Fore.GREEN}[OpenAI][3] {Fore.WHITE}Error: " + Fore.RED + "You have been rate limited")
                raise Exceptions.PyChatGPTException("You have been rate limited.")

            print(f"{Fore.GREEN}[OpenAI][3] {Fore.WHITE}Callback URL: {url}")
            self._part_four(url=url)
        elif response.status_code == 400:
            raise Exceptions.IPAddressRateLimitException("[OpenAI][3] Bad request from your IP address, "
                                                         "try again in a few minutes")
        else:
            raise Exceptions.Auth0Exception("[OpenAI][3] Failed to make the request, Try that again!")

    def _part_four(self, url: str):
        """
        We make a GET request to url
        :param url:
        :return:
        """
        headers = {
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Accept-Language': 'en-US,en;q=0.9',
            'Referer': 'https://chat.openai.com/',
        }
        print(f"{Fore.GREEN}[OpenAI][4] {Fore.WHITE}Making request to {url}")
        response = self.__session.get(url=url, headers=headers)
        if response.status_code == 302:
            print(f"{Fore.GREEN}[OpenAI][4] {Fore.WHITE}Request was " + Fore.GREEN + "successful")
            state = re.findall(r"state=(.*)", response.text)[0]
            state = state.split('"')[0]
            print(f"{Fore.GREEN}[OpenAI][4] {Fore.WHITE}Current State: {state}")
            self._part_five(state=state)
        else:
            raise Exceptions.Auth0Exception("[OpenAI][4] Failed to make the request, Try that again!")

    def _part_five(self, state: str):
        """
        We use the state to get the login page & check for a captcha
        """
        url = f"https://auth0.openai.com/u/login/identifier?state={state}"

        headers = {
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Accept-Language': 'en-US,en;q=0.9',
            'Referer': 'https://chat.openai.com/',
        }
        print(f"{Fore.GREEN}[OpenAI][5] {Fore.WHITE}Making request to {url}")
        response = self.__session.get(url, headers=headers)
        if response.status_code == 200:
            print(f"{Fore.GREEN}[OpenAI][5] {Fore.WHITE}Request was " + Fore.GREEN + "successful")
            soup = BeautifulSoup(response.text, 'lxml')
            if soup.find('img', alt='captcha'):
                print(f"{Fore.RED}[OpenAI][5] {Fore.RED}Captcha detected")

                svg_captcha = soup.find('img', alt='captcha')['src'].split(',')[1]
                decoded_svg = base64.decodebytes(svg_captcha.encode("ascii"))

                # Convert decoded svg to png
                drawing = svg2rlg(BytesIO(decoded_svg))

                # Better quality
                renderPM.drawToFile(drawing, "captcha.png", fmt="PNG", dpi=300)
                print(f"{Fore.GREEN}[OpenAI][5] {Fore.WHITE}Captcha saved to {Fore.GREEN}captcha.png"
                      + f" {Fore.WHITE}in the current directory")

                # Wait.
                captcha_input = input(f"{Fore.GREEN}[OpenAI][5] {Fore.WHITE}Please solve the captcha and "
                                      f"press enter to continue: ")
                if captcha_input:
                    print(f"{Fore.GREEN}[OpenAI][5] {Fore.WHITE}Continuing...")
                    self._part_six(state=state, captcha=captcha_input)
                else:
                    raise Exceptions.PyChatGPTException("[OpenAI][5] You didn't enter anything.")

            else:
                print(f"{Fore.GREEN}[OpenAI][5] {Fore.GREEN}No captcha detected")
                self._part_six(state=state, captcha=None)
        else:
            raise Exceptions.Auth0Exception("[OpenAI][5] Failed to make the request, Try that again!")

    def _part_six(self, state: str, captcha: str or None):
        """
        We make a POST request to the login page with the captcha, email
        :param state:
        :param captcha:
        :return:
        """
        print(f"{Fore.GREEN}[OpenAI][6] {Fore.WHITE}Making request to https://auth0.openai.com/u/login/identifier")
        url = f"https://auth0.openai.com/u/login/identifier?state={state}"
        email_url_encoded = self._url_encode(self.email_address)
        payload = f'state={state}&username={email_url_encoded}&captcha={captcha}&js-available=true&webauthn-available=true&is-brave=false&webauthn-platform-available=true&action=default'

        if captcha is None:
            payload = f'state={state}&username={email_url_encoded}&js-available=false&webauthn-available=true&is-brave=false&webauthn-platform-available=true&action=default'

        headers = {
            'Host': 'auth0.openai.com',
            'Origin': 'https://auth0.openai.com',
            'Connection': 'keep-alive',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Referer': f'https://auth0.openai.com/u/login/identifier?state={state}',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        response = self.__session.post(url, headers=headers, data=payload)
        if response.status_code == 302:
            print(f"{Fore.GREEN}[OpenAI][6] {Fore.WHITE}Email found")
            self._part_seven(state=state)
        else:
            raise Exceptions.Auth0Exception("[OpenAI][6] Email not found, Check your email address and try again!")

    def _part_seven(self, state: str):
        """
        We enter the password
        :param state:
        :return:
        """
        print(f"{Fore.GREEN}[OpenAI][7] {Fore.WHITE}Entering password...")
        url = f"https://auth0.openai.com/u/login/password?state={state}"

        email_url_encoded = self._url_encode(self.email_address)
        password_url_encoded = self._url_encode(self.password)
        payload = f'state={state}&username={email_url_encoded}&password={password_url_encoded}&action=default'
        headers = {
            'Host': 'auth0.openai.com',
            'Origin': 'https://auth0.openai.com',
            'Connection': 'keep-alive',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Referer': f'https://auth0.openai.com/u/login/password?state={state}',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        response = self.__session.post(url, headers=headers, data=payload)
        is_302 = response.status_code == 302
        if is_302:
            print(f"{Fore.GREEN}[OpenAI][7] {Fore.WHITE}Password was " + Fore.GREEN + "correct")
            new_state = re.findall(r"state=(.*)", response.text)[0]
            new_state = new_state.split('"')[0]
            print(f"{Fore.GREEN}[OpenAI][7] {Fore.WHITE}Old state: {Fore.GREEN}{state}")
            print(f"{Fore.GREEN}[OpenAI][7] {Fore.WHITE}New State: {Fore.GREEN}{new_state}")
            self._part_eight(old_state=state, new_state=new_state)
        else:
            raise Exceptions.Auth0Exception("[OpenAI][7] Password was incorrect or captcha was wrong")

    def _part_eight(self, old_state: str, new_state):
        url = f"https://auth0.openai.com/authorize/resume?state={new_state}"
        print(f"{Fore.GREEN}[OpenAI][8] {Fore.WHITE}Making request to {Fore.GREEN}{url}")
        headers = {
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15',
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Referer': f'https://auth0.openai.com/u/login/password?state={old_state}',
        }
        response = self.__session.get(url, headers=headers, allow_redirects=True)
        is_200 = response.status_code == 200
        if is_200:
            print(f"{Fore.GREEN}[OpenAI][8] {Fore.WHITE}All good")
            soup = BeautifulSoup(response.text, 'lxml')
            # Find __NEXT_DATA__, which contains the data we need, the get accessToken
            next_data = soup.find("script", {"id": "__NEXT_DATA__"})
            # Access Token
            access_token = re.findall(r"accessToken\":\"(.*)\"", next_data.text)
            if access_token:
                access_token = access_token[0]
                access_token = access_token.split('"')[0]
                print(f"{Fore.GREEN}[OpenAI][8] {Fore.WHITE}Access Token: {Fore.GREEN}{access_token}")
                # Save access_token
                self.save_access_token(access_token=access_token)
            else:
                print(f"{Fore.GREEN}[OpenAI][8][CRITICAL] {Fore.WHITE}Access Token: {Fore.RED}Not found"
                      f" Auth0 did not issue an access token.")
                self.part_nine()

    def part_nine(self):
        print(f"{Fore.GREEN}[OpenAI][9] {Fore.WHITE}"
              f"Attempting to get access token from: https://chat.openai.com/api/auth/session")
        url = "https://chat.openai.com/api/auth/session"
        headers = {
            "Host": "ask.openai.com",
            "Connection": "keep-alive",
            "If-None-Match": "\"bwc9mymkdm2\"",
            "Accept": "*/*",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15",
            "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
            "Referer": "https://chat.openai.com/chat",
            "Accept-Encoding": "gzip, deflate, br",
        }
        response = self.__session.get(url, headers=headers)
        is_200 = response.status_code == 200
        if is_200:
            print(f"{Fore.GREEN}[OpenAI][9] {Fore.GREEN}Request was successful")
            if 'json' in response.headers['Content-Type']:
                json_response = response.json()
                access_token = json_response['accessToken']
                print(f"{Fore.GREEN}[OpenAI][9] {Fore.WHITE}Access Token: {Fore.GREEN}{access_token}")
                self.save_access_token(access_token=access_token)
            else:
                print(f"{Fore.GREEN}[OpenAI][9] {Fore.WHITE}Access Token: {Fore.RED}Not found, "
                      f"Please try again with a proxy (or use a new proxy if you are using one)")
        else:
            print(f"{Fore.GREEN}[OpenAI][9] {Fore.WHITE}Access Token: {Fore.RED}Not found, "
                  f"Please try again with a proxy (or use a new proxy if you are using one)")

    # @staticmethod
    def save_access_token(self, access_token: str, expiry: int or None = None):
        """
        Save access_token and an hour from now on CHATGPT_ACCESS_TOKEN CHATGPT_ACCESS_TOKEN_EXPIRY environment variables
        :param expiry:
        :param access_token:
        :return:
        """
        try:
            save_auth_path = f"{self.email_address}_auth.json"
            print(f"{Fore.GREEN}[OpenAI][9] {Fore.WHITE}Saving access token to {save_auth_path}...")
            expiry = expiry or int(time.time()) + 3600

            # Get path using os, it's in ./classes/auth.json
            path = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(path, save_auth_path)
            with open(path, "w") as f:
                f.write(json.dumps({"access_token": access_token, "expires_at": expiry}))

            print(f"{Fore.GREEN}[OpenAI][8] {Fore.WHITE}Saved access token for {self.email_address}")
        except Exception as e:
            raise e



# adapted from https://github.com/rawandahmad698/PyChatGPT

# Builtins
import sys
import time
import os
from queue import Queue
import json
from typing import Tuple

# Local
# from pychatgpt.classes import openai as OpenAI
from pychatgpt.classes import chat as ChatHandler
from pychatgpt.classes import exceptions as Exceptions
from pychatgpt.main import Options
from pychatgpt.main import Chat
from .auth import Auth

# Fancy stuff
from colorama import Fore


def token_expired(email_address) -> bool:
    f"""
        Check if the creds have expired for {email_address}
        returns:
            bool: True if expired, False if not
    """
    try:
        # Get path using os, it's in ./classes/auth.json
        path = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(path, f"{email_address}_auth.json")

        with open(path, 'r') as f:
            creds = json.load(f)
        expires_at = float(creds['expires_at'])
        if time.time() > expires_at + 3600:
            return True
        else:
            return False
    except KeyError:
        return True
    except FileNotFoundError:
        return True


def get_access_token(email_address) -> Tuple[str or None, str or None]:
    """
        Get the access token for {email_address}
        returns:
            str: The access token
    """
    try:
        # Get path using os, it's in ./Classes/auth.json
        path = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(path, f"{email_address}_auth.json")

        with open(path, 'r') as f:
            creds = json.load(f)
            return creds['access_token'], creds['expires_at']
    except FileNotFoundError:
        return None, None


class FastChat(Chat):
    def __init__(self,
                 email: str,
                 password: str,
                 options: Options or None = None,
                 conversation_id: str or None = None,
                 previous_convo_id: str or None = None):
        self.email = email
        self.password = password
        self.options = options

        self.conversation_id = conversation_id
        self.previous_convo_id = previous_convo_id

        self.auth_access_token: str or None = None
        self.auth_access_token_expiry: int or None = None
        self.__chat_history: list or None = None

        self._setup()

    @staticmethod
    def _create_if_not_exists(file: str):
        if not os.path.exists(file):
            with open(file, 'w') as f:
                f.write("")

    def _setup(self):
        if self.options is not None:
            # If track is enabled, create the chat log and id log files if they don't exist
            if not isinstance(self.options.track, bool):
                raise Exceptions.PyChatGPTException("Options to track conversation must be a boolean.")

            if self.options.track:
                if self.options.chat_log is not None:
                    self._create_if_not_exists(self.options.chat_log)
                    self.options.id_log = os.path.abspath(self.options.chat_log)
                else:
                    # Create a chat log file called chat_log.txt
                    self.options.chat_log = "chat_log.txt"
                    self._create_if_not_exists(self.options.chat_log)

                if self.options.id_log is not None:
                    self._create_if_not_exists(self.options.id_log)
                    self.options.id_log = os.path.abspath(self.options.id_log)
                else:
                    # Create a chat log file called id_log.txt
                    self.options.id_log = "id_log.txt"
                    self._create_if_not_exists(self.options.id_log)

            if self.options.proxies is not None:
                if not isinstance(self.options.proxies, dict):
                    if not isinstance(self.options.proxies, str):
                        raise Exceptions.PyChatGPTException("Proxies must be a string or dictionary.")
                    else:
                        self.proxies = {"http": self.options.proxies, "https": self.options.proxies}
                        print(f"{Fore.GREEN}>> Using proxies: True.")

            if self.options.track:
                print(f"{Fore.GREEN}>> Tracking conversation enabled.")
                if not isinstance(self.options.chat_log, str) or not isinstance(self.options.id_log, str):
                    raise Exceptions.PyChatGPTException(
                        "When saving a chat, file paths for chat_log and id_log must be strings.")
                elif len(self.options.chat_log) == 0 or len(self.options.id_log) == 0:
                    raise Exceptions.PyChatGPTException(
                        "When saving a chat, file paths for chat_log and id_log cannot be empty.")

                self.__chat_history = []
        else:
            self.options = Options()

        if not self.email or not self.password:
            print(f"{Fore.RED}>> You must provide an email and password when initializing the class.")
            raise Exceptions.PyChatGPTException("You must provide an email and password when initializing the class.")

        if not isinstance(self.email, str) or not isinstance(self.password, str):
            print(f"{Fore.RED}>> Email and password must be strings.")
            raise Exceptions.PyChatGPTException("Email and password must be strings.")

        if len(self.email) == 0 or len(self.password) == 0:
            print(f"{Fore.RED}>> Email cannot be empty.")
            raise Exceptions.PyChatGPTException("Email cannot be empty.")

        if self.options is not None and self.options.track:
            try:
                with open(self.options.id_log, "r") as f:
                    # Check if there's any data in the file
                    if len(f.read()) > 0:
                        self.previous_convo_id = f.readline().strip()
                        self.conversation_id = f.readline().strip()
                    else:
                        self.conversation_id = None

            except IOError:
                raise Exceptions.PyChatGPTException("When resuming a chat, conversation id and previous conversation id in id_log must be separated by newlines.")
            except Exception:
                raise Exceptions.PyChatGPTException("When resuming a chat, there was an issue reading id_log, make sure that it is formatted correctly.")

        # Check for access_token & access_token_expiry in env
        if token_expired(self.email):
            print(f"{Fore.RED}>> Access Token missing or expired for {self.email}."
                  f" {Fore.GREEN}Attempting to create them...")
            self._create_access_token()
        access_token, expiry = get_access_token(self.email)
        self.auth_access_token = access_token
        self.auth_access_token_expiry = expiry

        try:
            self.auth_access_token_expiry = int(self.auth_access_token_expiry)
        except ValueError:
            print(f"{Fore.RED}>> Expiry is not an integer.")
            raise Exceptions.PyChatGPTException("Expiry is not an integer.")

        if self.auth_access_token_expiry < time.time():
            print(f"{Fore.RED}>> Your access token is expired for {self.email}. {Fore.GREEN}Attempting to recreate it...")
            self._create_access_token()

    def _create_access_token(self) -> bool:
        openai_auth = Auth(email_address=self.email, password=self.password, proxy=self.options.proxies)
        openai_auth.create_token()

        # If after creating the token, it's still expired, then something went wrong.
        is_still_expired = token_expired(self.email)
        if is_still_expired:
            print(f"{Fore.RED}>> Failed to create access token for {self.email}.")
            return False

        # If created, then return True
        return True

    def ask(self, prompt: str, rep_queue: Queue or None = None) -> str or None:
        if prompt is None:
            print(f"{Fore.RED}>> Enter a prompt.")
            raise Exceptions.PyChatGPTException("Enter a prompt.")

        if not isinstance(prompt, str):
            raise Exceptions.PyChatGPTException("Prompt must be a string.")

        if len(prompt) == 0:
            raise Exceptions.PyChatGPTException("Prompt cannot be empty.")

        if rep_queue is not None and not isinstance(rep_queue, Queue):
            raise Exceptions.PyChatGPTException("Cannot enter a non-queue object as the response queue for threads.")

        # Check if the access token is expired
        if token_expired(self.email):
            print(f"{Fore.RED}>> Your access token for {self.email} is expired. {Fore.GREEN}Attempting to recreate it...")
            did_create = self._create_access_token()
            if did_create:
                print(f"{Fore.GREEN}>> Successfully recreated access token.")
            else:
                print(f"{Fore.RED}>> Failed to recreate access token for {self.email}.")
                raise Exceptions.PyChatGPTException(f"Failed to recreate access token for {self.email}.")

        # Get access token
        access_token = get_access_token(self.email)

        answer, previous_convo, convo_id = ChatHandler.ask(auth_token=access_token,
                                                           prompt=prompt,
                                                           conversation_id=self.conversation_id,
                                                           previous_convo_id=self.previous_convo_id,
                                                           proxies=self.options.proxies)

        if rep_queue is not None:
            rep_queue.put((prompt, answer))

        if answer == "400" or answer == "401":
            print(f"{Fore.RED}>> Failed to get a response from the API with {self.email}.")
            return None

        self.conversation_id = convo_id
        self.previous_convo_id = previous_convo

        if self.options.track:
            self.__chat_history.append("You: " + prompt)
            self.__chat_history.append("Chat GPT: " + answer)
            self.save_data()

        return answer

    def save_data(self):
        if self.options.track:
            try:
                with open(self.options.chat_log, "a") as f:
                    f.write("\n".join(self.__chat_history) + "\n")

                with open(self.options.id_log, "w") as f:
                    f.write(str(self.previous_convo_id) + "\n")
                    f.write(str(self.conversation_id) + "\n")

            except Exception as ex:
                print(f"{Fore.RED}>> Failed to save chat and ids to chat log and id_log."
                      f"{ex}")
            finally:
                self.__chat_history = []



import threading
import time


class ShowTyping(threading.Thread):
    """
    Use to show the waiting information

    """
    def __init__(self, prompt, show_time=True, update_every=0.3):
        super().__init__()
        self.show_time = show_time
        self.update_every = update_every
        self.setDaemon(True)
        self.print_typing = False
        self.need_clear = False
        self.prompt = prompt

    def run(self):
        while True:
            text = self.prompt + "I'm thinking, please wait"
            start_time = time.time()
            count = 0
            _text = ''
            while self.print_typing:
                self.need_clear = True
                if count>10:
                    print(text+' '*16, end='\r')
                    count = 0
                _text = text + '.'*count
                count += 1
                if self.show_time:
                    seconds = round(time.time() - start_time, 1)
                    _text += f'{seconds}s'
                print(_text, end='\r')
                time.sleep(self.update_every)
            if count != 0:
                print(' '*len(_text), end='\r')
                print(self.prompt, end='', flush=True)
            self.need_clear = False
            time.sleep(0.01)

    def start_waiting(self):
        self.print_typing = True
        self.need_clear = True

    def end_waiting(self):
        self.print_typing = False
        count = 0
        while self.need_clear:  # waiting for clearing
            count += 1
            time.sleep(0.01)
            if count > 2*(self.update_every/0.01):
                print("There is a bug....")
                break



import argparse
import os
import time
import uuid

from colorama import Style, Fore
import logging
from logging import FileHandler, Formatter, NullHandler
import types
from datetime import datetime

from tls_client.exceptions import TLSClientExeption

from fastchatgpt.utils.show_typing import ShowTyping
from pychatgpt import Chat

parser = argparse.ArgumentParser()


def bot_play():
    parser.add_argument("--bot", type=str, default="ChatGPT")
    parser.add_argument('--username', type=str,  required=True, help='Username to login in')
    parser.add_argument('--password', type=str, required=True, help='Password to login in')
    parser.add_argument('--proxies', type=str, default='', help='Proxy to use')
    parser.add_argument('--output', type=str, default='play_logs.txt', help='Where to save the dialogue history')

    args = parser.parse_known_args()[0]

    proxies = args.proxies
    if proxies is not None:
        assert isinstance(proxies, str), 'Only string proxies information is allowed. ' \
                                         'Such as http_proxy=http://proxy.com+https_proxy=http://proxy.com.'
        parts = proxies.split('+')
        _proxies = {}
        for part in parts:
            if part.startswith('http_proxy='):
                _proxies['http'] = part.split('http_proxy=')[1]
            if part.startswith('https_proxy='):
                _proxies['https'] = part.split('https_proxy=')[1]

    print("Start to log in openai...")
    from pychatgpt import Options
    options = Options()
    options.proxies = proxies
    bot = Chat(email=args.username, password=args.password, options=options)
    logger = logging.getLogger('fastchatgpt')
    if args.output is not None or args.output != 'None':
        fp = os.path.abspath(args.output)
        dirname = os.path.dirname(fp)
        os.makedirs(dirname, exist_ok=True)

        logger.setLevel(logging.DEBUG)
        handler = FileHandler(fp, mode='a', encoding='utf8')
        formatter = Formatter(fmt='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        logger.addHandler(handler)
        logger.propagate = False  # forbid pass log to upper logger
        # for separator and some other info to log
        def log_pure_message(self, message):
            # Switch formatter, output a blank line
            self.handler.setFormatter(self.blank_formatter)
            self.info(message)
            # Switch back
            self.handler.setFormatter(self.formatter)
        logger.handler = handler
        blank_formatter = logging.Formatter(fmt="")
        logger.formatter = formatter
        logger.blank_formatter = blank_formatter
        logger.pure_info = types.MethodType(log_pure_message, logger)

        print(Fore.GREEN+f'play logs will be saved in {fp}.'+Style.RESET_ALL)
        logger.pure_info(f'\n\nNew play logs on {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n')
    else:
        logger.addHandler(NullHandler())

    # print info
    print("bot_play - Help user to play with chat bot and log the conversation")
    print("Please use \\n to represent change line in your input")
    help_msg = """
                !help - Show help message
                !sep {message} - Add a sep message to the log file
                !exit - Exit the program
                !refresh - Forget current dialogue history
    """
    print(help_msg)
    bot_prompt = Fore.BLUE+"Bot: "
    show_type = ShowTyping(bot_prompt)
    show_type.start()
    start_time = time.time()
    num_turns = 0
    response_total_time = 0
    while True:
        # Here, we need to use color to make the interaction more pretty
        prompt = 'You: '
        # text = get_input(prompt)
        text = input(prompt)

        if text == '!help':
            print(help_msg)
            continue
        elif text.startswith('!sep'):
            info = text.split('!sep')[1].strip()
            info = '='*10 + info + '='*10
            logger.pure_info('\n\n' + info + '\n\n')
            continue
        elif text.startswith('!quit') or text.startswith('!exit'):
            total_time = round(time.time() - start_time, 2)
            response_avg_time = round(response_total_time/(num_turns+1e-6), 2)
            print(Fore.RED+f"Communicate with Bot {total_time} seconds, every turn takes "
                           f"Bot {response_avg_time} seconds to response."+Style.RESET_ALL)
            break
        elif text.startswith('!refresh'):
            bot.conversation_id = None
            bot.previous_convo_id = str(uuid.uuid4())
            continue

        logger.info(prompt + text)
        # print(bot_prompt, end="", flush=True)
        turn_output = ''
        turn_output += bot_prompt.split(Fore.BLUE)[1]
        show_type.start_waiting()
        retry = 0
        turn_start_time = time.time()
        while True:
            try:
                response = bot.ask(text)
                show_type.end_waiting()
                print(response)
                turn_output = response
                # response = next(bot.request(text))
                # show_type.end_waiting()
                # words = response["message"]
                # for word in words.split():
                #     print(word, end=' ', flush=True)
                #     turn_output += word + ' '
                #     time.sleep(random.randint(100, 300)/1000.0)
                retry += 1
                break
            except TLSClientExeption as e:
                if retry>3:
                    break
                raise e
        response_total_time += time.time() - turn_start_time
        num_turns += 1
        print(Style.RESET_ALL, flush=True)
        print()
        logger.info(turn_output)  # save the output from the bot
        logger.pure_info('')  # keep one empty line in the log

